// module hu_ftp2.d
// Simple file transfer via LAN for binary files (no code conversion)
// (not compatible with standard FTP!)
// Server and client are in the same CLI program
//
// Private Copyright Dr.E.Huckert 02-2018 

// compile:
//   dmd hu_ftp2.d d_trace.d srvsocket.d clientsocket.d msgsocket.d

// Protocols:
// g (= get file, from client perspective):
//   client: gfilename
//   server: file portion with length (length < 0 if error) 
//   client: "OK" or "ERR" as ack on each portion until length of portion <= 0
// Aufbau eines typischen Pakets:
//   - 6 Bytes feste Laenge (ASCII), z.B. "000006"
//   - 1 Byte Operationskennung, z.B. 'g' oder 'p'
//   - n Bytes Nutzdaten
//   - die Laenge umfasst Operationskennung  plus Nutzdaten!
// Typische Quittung (ACK):
//   "000003gOK"
// Note: in order to distguish ths version 2 from version 1  the
//       operatio bytes are now lower case letters.
//       Both protocol versions are not(!) compatible
// Versions:
// 13.02.2018 first version (modeled after my program hu_ftp.d)

// V1.X27

import std.stdio;
import std.string;
import std.conv;
import std.datetime;
import core.thread;
import core.time;
import core.stdc.stdio;
import core.stdc.string;
import core.stdc.stdlib;
import std.socket;
import std.file;  // for exists()

import srvsocket;
import msgsocket;
import clientsocket;
import d_trace;

immutable int SERVER = 1;
immutable int CLIENT = 0;
immutable int RETRY_MAX = 100;
 
// --------------- globals ----------------------------
int      traceLevel  = 0;
D_TRACE  dTrace      = null;
// ----------------end of globals ---------------------

// ----------------------------------------------------
// Send an acknowledge 
int send_ack(CLIENTSOCKET clientSock, int kind)
{
  int       ret = 0;
  char []   cmd;
  MSGSOCKET msgSock = null;
  //
  if (kind == 1)      
    cmd = "ok".dup;
  else 
    cmd = "err".dup;
  char opByte = 'p';
  if (kind == 0)
      opByte = 'e';
  try
  {
    msgSock    = new MSGSOCKET(opByte, cast(ubyte[])cmd);
    ret  = clientSock.send(msgSock, 0);
    if (dTrace !is null)
      dTrace.trace("CL: send_ack(): ACK sent ret=", ret);
    delete msgSock;
  }
  catch (SocketException e)
  {
    writeln("SRVSOCKET EXCEPTION: socket (send_ack())");
    ret = -1;
  }
  //
  zurueck:
  return ret;
};   // end send_ack()

// ----------------------------------------------------
// Receive an acknowledge from the client
// Returns an allocate MSGSOCKET instance that must be freed externally
MSGSOCKET receive_ack(SRVSOCKET clientSock)
{
  int       ret = 0;
  char []   cmd;
  MSGSOCKET msgSock = null;
  //
  msgSock = clientSock.recv();
  if (dTrace !is null)
  {
    string msg = cast(string)(msgSock.get_content());
    dTrace.trace("SR: receive_ack(): ACK received:", msg);
  }
  return msgSock;
};   // end receive_ack()

// -----------------------------------------------------
// Server: Read a binary file and send it via the socket
// Returns: >= 0 if ok
//          < 0  if error
int send_file(SRVSOCKET srvSock,
              char [] filename)
{
  MSGSOCKET msgSock = null;
  MSGSOCKET ackMsg  = null;
  int   ret = 0;
  std.stdio.File  ifd; // never use this:  = null;
  bool eofFlag = false;
  int   nb;
  uint  noPack = 0;
  const uint BUFLENGTH = 2048;
  ubyte [BUFLENGTH] buf2;  // 2KB read buffer
  ubyte [] buf1;
  string msg;
  ulong  noFileBytes = 0;
  MonoTime      t1, t2;
  //
  if (dTrace !is null)
    dTrace.trace("SR(D): send_file() filename=", filename);
  t1 = MonoTime.currTime;
  //
  // open the input file
  try
  {
    if (dTrace !is null)
      dTrace.trace("SR(D): send_file() opening file");
    // test if the file exists
    if (! exists(filename))
    {
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file(): file not found: ",
                     filename);
      writeln("SR(D): ERROR file not found: ", filename);
      msg = "ERROR: file not found";
      msgSock = new MSGSOCKET('e', cast(ubyte[])msg, cast(uint)msg.length);
      ret = srvSock.send(msgSock, 0);
      ret = -1;
      goto zurueck;
    }
    // open the file
    ifd = File(filename, "rb");
  }
  catch (StdioException e)
  {
    // should not occur
    if (dTrace !is null)
      dTrace.trace("SR(D): send_file(): stdio EXCEPTION: cannot open file=",
                    filename);
    writeln("SR(D): ERROR cannot open file: ", filename);
    msg = "ERROR: cannot open file";
    msgSock = new MSGSOCKET('e', cast(ubyte[])msg, cast(uint)msg.length);
    ret = srvSock.send(msgSock, 0);
    ret = -2;
    goto zurueck;
  }
  catch (FileException e)
  {
    if (dTrace !is null)
      dTrace.trace("SR(D): send_file(): file EXCEPTION: ",
                    filename);
    writeln("SR(D): file EXCEPTION: ", filename);
    msg = "file EXCEPTION";
    msgSock = new MSGSOCKET('e', cast(ubyte[])msg, cast(uint)msg.length);
    ret = srvSock.send(msgSock, 0);
    ret = -3;
    goto zurueck;
  }
  //
  srvSock.setNettoTx(0L);
  srvSock.setNettoRx(0L);
  //
  if (dTrace !is null)
      dTrace.trace("SR(D): send_file(): entering file read loop");
  while (! eofFlag)
  {
    // read next portion from file
    try
    {
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file() starting package=", noPack);
      // TODO: not correct - must be done in a loop!!!
      buf1 = ifd.rawRead(buf2);
      if (buf1.length < BUFLENGTH)
        eofFlag = true;
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file(): file buffer length=", buf1.length);
      nb   = cast(int)buf1.length;
    }
    catch(Exception e)
    {
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file(): file EXCEPTION");
      ifd.close();
      ret = -4;
      goto zurueck;
    }
    //
    // send now this portion as a 'p' package
    if (nb <= 0)
    {
      nb = 0;
      eofFlag = true;
    }
    msgSock = new MSGSOCKET('p', buf1, cast(uint)nb);
    ret = srvSock.send(msgSock, 0);
    if (dTrace !is null)
    {
      dTrace.trace("SR(D): send_file(): srvSock.send() ret=", ret);
      dTrace.trace("SR(D): send_file() TX packet no=", noPack); 
      dTrace.trace("SR(D): send_file() netto bytes=", nb);
    } 
    if (ret != (msgSock.get_content_length() + 6 + 1))
    {
      writeln("SR(D): ERROR Not all bytes sent");
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file(): ERROR Not all bytes sent");
    }
    //
    // receive a positive ACK("ok") for each package
    if (dTrace !is null)
      dTrace.trace("SR(D): send_file() waiting for Ack"); 
    ackMsg = receive_ack(srvSock);
    msg = cast(string)(ackMsg.get_content());
    if (msg != "ok")
    {
      if (dTrace !is null)
        dTrace.trace("SR(D): send_file() ACK error: ", msg);
      ret = -7;
      goto zurueck;
    }
    noFileBytes += nb;
    noPack++;
    delete ackMsg;
    ackMsg = null;
    delete msgSock;
    msgSock = null;
  }   // end while (! eofFlag)
  //
  // Send the last (=EOF) package with content length = 0
  if (dTrace !is null)
    dTrace.trace("SR(D): send_file() starting EOF package=", noPack);
  buf1.length = 0;
  msgSock = new MSGSOCKET('p', buf1, 0);
  ret = srvSock.send(msgSock, 0);
  if (dTrace !is null)
  {
    dTrace.trace("SR(D): send_file(): EOF srvSock.send() ret=", ret);
    dTrace.trace("SR(D): send_file() EOF TX packet no=", noPack); 
  } 
  noPack++;
  //
  // OK here
  ret = noPack;
  //
  zurueck:
  if (msgSock !is null)
    delete msgSock;
  if (ackMsg !is null)
    delete ackMsg;
  if (dTrace !is null)
    dTrace.trace("SR(D): send_file() netto TX=", srvSock.getNettoTx());
  t2 = MonoTime.currTime;
  Duration t = t2 - t1;
  writeln("Time consumed(ms):", t.total!"msecs");
  if (ret >= 0) 
  {
    if (ifd.size == noFileBytes)
      writeln("SR(D): send_file() File completely sent: ", filename);
    else
      writeln("SR(D): send_file() Warning: file uncompletely sent: ", filename);
    ifd.close();
    if (dTrace !is null)
    {
      dTrace.trace("SR(D): send_file() time in ms: ", t.total!"msecs");
      dTrace.trace("SR(D): send_file() ret=", ret);
    }
  }
  return ret;
}   // end send_file()

// -----------------------------------------------------------
int send_request(CLIENTSOCKET sock,
                 char op, 
                 const string fileName)
{
  int ret = 0;
  MSGSOCKET msgSock  = new MSGSOCKET(' ', "");
  //
  if (dTrace !is null)
    dTrace.trace("CL(D): send_request() fileName=", fileName);
  msgSock.set_operation(op);
  msgSock.set_content(cast(ubyte[])fileName);
  msgSock.set_content_length(cast(uint)(fileName.length));
  //
  ret = sock.send(msgSock, 0);
  //
  zurueck:
  delete msgSock;
  if (dTrace !is null)
    dTrace.trace("CL(D): send_request() ret=", ret);
  return ret;
}  // end send_request()

// -----------------------------------------------------------
// Do the client work: request a file and receive it
// returns: >=0 if ok
//          < 0 if error
int do_client(const string hostname,
              int port,
              const string filename)
{
  MSGSOCKET     msgSock  = null;
  CLIENTSOCKET  clientSock = null;
  int           ret    = 0;
  File          ofd;        // never use this:  = null;
  int           ll, nb = 0;
  char   []     cmd;
  int           anz, cRet;
  MonoTime      t1, t2;
  char          *caddr = null;
  ubyte []      ubyteBuf;
  uint          noPack = 0;
  bool          eofFlag = false;
  //
  if (dTrace !is null)
    dTrace.trace("CL(D): do_client(): start");
  t1 = MonoTime.currTime;
  //
  // open a socket connection to a server 
  // The server must be started
  DateTime time1 = cast(DateTime)(Clock.currTime());
  //
  clientSock = new CLIENTSOCKET();
  clientSock.setLogging(dTrace, traceLevel);
  clientSock.setNettoTx(0L);
  clientSock.setNettoRx(0L);
  //
  ret   = clientSock.open(hostname, port);
  if (ret < 0)
  {
    writeln("CL(D): do_client(): ERROR cannot open the socket: ",
            std.socket.lastSocketError());
    if (dTrace !is null)
      dTrace.trace("CL(D): do_client(): ERROR cannot open the socket: ",
            std.socket.lastSocketError());
    ret = -1 ;
    goto zurueck;
  }
  if ((dTrace !is null) && (traceLevel > 1))
    dTrace.trace("CL(D): do_client(): socket open() ret=", ret);
  //
  // send a "get" request for a file
  ret = send_request(clientSock, 'g', filename);
  if (dTrace !is null)
    dTrace.trace("CL(D): do_client() send_request() ret=", ret);
  if (ret < 0)
  {
    ret = -2;
    goto zurueck;
  }
  // request for file has been sent
  // Wait here for content packages, "err" package or 
  // EOF-package (empty content)
  while (! eofFlag)
  {
    // receive the next package from the server 
    if (dTrace !is null)
      dTrace.trace("CL(D): do_client(): rx next package no=", noPack);
    msgSock = clientSock.recv();
    if (msgSock.get_operation() == 'e')
    {
      // error package received
      if (dTrace !is null)
        dTrace.trace("CL(D): do_client() ERROR signalled by server");
      string srvErrMsg = cast(string)(msgSock.get_content());
      if (dTrace !is null)
        dTrace.trace("CL(D): do_client() ERROR msg from server=",
                     srvErrMsg);
      writeln("Error message from server: ", srvErrMsg);
      // 
      ret = -4;
      goto zurueck;
    }
    if (msgSock.get_operation() != 'e')
    {
      // this is not an error package (true content exists)
      if (noPack == 0)
      {
        // first content package: open the output file
        try
        {
          ofd = File(filename, "wb");
        }
        catch (StdioException e)
        {
          writeln("CL(D): do_client() ERROR cannot open local file:",  cmd);
          dTrace.trace("CL(D): do_client() ERROR: cannot open local file:");
          ret = -5;
          goto zurueck;
        }
        if (dTrace !is null)
          dTrace.trace("CL(D): do_client(): outp. file is open");
      } // if noPack == 0;
      //
      // ll = netto length (without the operation byte!)
      ll = cast(int)(msgSock.get_content_length());
      if (dTrace !is null)
        dTrace.trace("CL(D): do_client() RX OK netto=", ll); 
      if (ll > 0)
      {
        // package is not empty
        // write the content to the file on disk
        ubyteBuf = msgSock.get_content();  // ohne Op.byte!
        try
        {
          ofd.rawWrite(ubyteBuf);
        }
        catch (Exception e)
        {
          // Schreibfehler
          if (dTrace !is null)
            dTrace.trace("CL(D): do_client() file write EXCEPTION");
          ret = -6;
          goto zurueck;
        }
      }   // end  if (ll > 0)
      else
      {
        if (dTrace !is null)
            dTrace.trace("CL(D): do_client() last package detected!");
        eofFlag = true;
      }
    }   // end if (msgSock.get_operation() != 'e')
    //
    // send acknowledge "OK"
    if (dTrace !is null)
      dTrace.trace("CL(D): do_client() sending ACK=\"ok\"");
    int locRet = send_ack(clientSock, 1); 
    if (locRet < 0)
    {
      writeln("do_client(): ERROR TCP send");
      ret = -5;
      goto zurueck;
    }
    noPack++;
    if (msgSock !is null)
    {
      delete msgSock;
      msgSock = null;
    }
  }   // end while (! eofFlag)
  //
  zurueck:
  if (clientSock !is null)
  {
    clientSock.close();
    delete clientSock;
  }
  if (msgSock !is null)
  {
    delete msgSock;
    msgSock = null;
  }
  t2 = MonoTime.currTime;
  Duration t = t2 - t1;
  writeln("CL(D): Time consumed(ms):", t.total!"msecs");
  if (dTrace !is null)
    dTrace.trace("CL(D): do_client(): ret=", ret);
  return ret;
}   // end do_client()

// -----------------------------------------------------------
// server for our file transfer
// returns: >=0 if ok
//          < 0 if error
int do_server(ushort port)
{
  SRVSOCKET     srvSock = null;
  MSGSOCKET     msgSock = null;
  char   [512]  answer;
  bool          end_flag = false;
  int           ret      = 0;
  int           ret_code = 0;
  char          operation;
  char []       request;
  int           retry_count = 0;
  //
  if (dTrace !is null)
    dTrace.trace("do_server() ---- start ----");
  //
  srvSock = new SRVSOCKET(); 
  srvSock.set_timeout(1000L); 
  srvSock.setLogging(dTrace, traceLevel);
  if (srvSock.open(port) < 0)
  {
    writeln("SR(D) do_server(): ERROR: cannot open socket: ",
             std.socket.lastSocketError());
    if (dTrace !is null)
      dTrace.trace("SR(D) do_server(): ERROR: cannot open socket: ",
                    std.socket.lastSocketError());
    ret_code = -1;
    goto zurueck;
  }
  if (dTrace !is null)
    dTrace.trace("SR(D) do_server(): listening...");
  ret = srvSock.listen(1);
  if (dTrace !is null)
    dTrace.trace("SR(D) do_server(): result listen()=", ret);
  if (ret < 0)
  {
    writeln("SRVSOCKET: error listen()",
             std.socket.lastSocketError());
    if (dTrace !is null)
      dTrace.trace("SR(D) do_server(): listen error: ",
                    std.socket.lastSocketError());
    ret_code = -2;
    goto zurueck;
  }
  //
  if (dTrace !is null)
    dTrace.trace("SR(D) do_server(): accepting...");
  ret = srvSock.accept();
  if (dTrace !is null)
    dTrace.trace("SR(D) do_server(): accept() ret=", ret);
  if (ret < 0)
  {
    writeln("SRVSOCKET: error accept(): ", 
             std.socket.lastSocketError());
    if (dTrace !is null)
      dTrace.trace("SR(D) do_server(): accept error: ",
                    std.socket.lastSocketError());
    ret_code = -3;
    goto zurueck;
  }
  srvSock.setNettoTx(0L);
  srvSock.setNettoRx(0L);
  //
  writeln("SR(D) do_server(): waiting for a file request...");
  if (dTrace !is null)
    dTrace.trace("SR(D) do_server(): waiting for file request...");
  end_flag = false;
  while (! end_flag)
  {
    if (! srvSock.testDataAvailable(100))
    {
      // Timeout - nothing yet received
      srvSock.milli_sleep(100);
      continue;
    }
    // here are data available
    // get a file name package from the client or
    // an exit package
    msgSock = srvSock.recv();
    if (msgSock.get_operation() == 'e')
    {
      string msg = cast(string)(msgSock.get_content());
      writeln ("SR(D) do_server(): MSGSOCKET instance error:",
               msg);
      if (dTrace !is null)
        dTrace.trace("SR(D) do_server(): MSGSOCKET instance error:",
               msg);
      break;
    }
    request   = cast(char [])(msgSock.get_content());
    operation = msgSock.get_operation();
    if (dTrace !is null)
    {
      string aux = "";
      aux ~= operation;
      dTrace.trace("SR(D) do_server(): operation=", aux);
      if (dTrace !is null)
        dTrace.trace("SR(D) do_server(): g request: file=", request);
    }
    //
    if (operation == 'g')
    {
      writeln("file request: ", request);
      ret = send_file(srvSock, request);
      if (dTrace !is null)
        dTrace.trace("SR(D) do_server(): result send_file=", ret);      
      end_flag = true; 
    }
    else
    {
      if (dTrace !is null)
        dTrace.trace("SR(D) do_server(): not a 'g' operation!");        
      end_flag = true; 
    }
  }   // end while (! end_flag)
  //
  zurueck:
  if (srvSock !is null)
    // delete invokes the destructor: this closes both server sockets
    delete srvSock;
  if (msgSock !is null)
    delete msgSock;
  return ret_code;
}   // end do_server()

// -----------------------------------------------------------
void usage()
{
  writeln("hu_ftp2 V1.X27 Copyright Dr.E.Huckert 2018-02" );
  writeln("usage: hu_ftp2 -port nnn [-host name] [-server|-client] [-v] [-v2] [-i filename]" );
}   // end usage()

// -----------------------------------------------------------
int main(string [] args)
{
  int       n, ret;  
  ushort    port          = 17777;
  int       role          = SERVER;
  string    hostname      = "localhost";
  string    filename      = "";
  //
  usage();
  //
  // Evaluate the command line arguments
  for (n=1; n < args.length; n++)
  {
    if (args[n] == "-port")
      port = parse!ushort(args[n+1]);
    if (args[n] ==  "-host")
      hostname = args[n+1];
    if (args[n] ==  "-server") 
      role = SERVER;
    if (args[n] ==  "-client")
      role = CLIENT;
    if (args[n] ==  "-v")
      traceLevel = 1;
    if (args[n] ==  "-v2")
      traceLevel = 2;
    if (args[n] ==  "-i")
      filename = args[n+1];
  }
  if (traceLevel > 0)
    dTrace = new D_TRACE("localhost", 9514);
  if (dTrace !is null)
  {
    dTrace.trace("----- new session ----- role= ", role);
    writeln("main(D): port=", port);
    dTrace.trace("main(D): port=", port);
  }
  if (role == SERVER)
  {
    // remains endless in server mode
    while (true)
    {
      ret = do_server(port);
      if (dTrace !is null)
        dTrace.trace("main(D): do_server() operation ret=", ret);
      if (ret < 0)
        break;
      // seems necessary under Linux
      Thread.sleep(dur!("msecs")(250));
    }
  }
  else
  {
    ret = do_client(hostname, port, filename);
    if (dTrace !is null)
       dTrace.trace("main(D): client operation ret=", ret);
  }
  return 0;
}   // end main()
